@@ -1,18 +1,19 @@
 import threading
-import random
 import asyncio
 import queue
 import os
 import time
-import sys
 
-from typing import Optional, TypeVar, Union, cast, List, Dict
+from functools import reduce
+from random import getrandbits, randrange
+from typing import Any, Callable, Coroutine, Optional, Tuple, TypeVar, Union, cast, List, Dict
+from textwrap import wrap
 
 from .constants import *
+from .logger.constants import *
 from .control.constants import *
 from .system.constants import *
 from .unique_brain.constants import *
-from .events import *
 from .packet import *
 
 from .util import now, log, logv
@@ -35,10 +36,132 @@ EV_IDENTIFY = "identify"
 EV_CONNECTED = "connected"
 EV_DISCONNECTED = "disconnected"
 
-#_ACK_RETRIES = const(4)
-#_ACK_DELAY = const(40)
+# _ACK_RETRIES = const(4)
+# _ACK_DELAY = const(40)
 
 RegType = TypeVar('RegType', bound=Union[PackType, PackTuple])
+HandlerFn = Callable[..., Union[None, Coroutine[Any, Any, None]]]
+EventHandlerFn = Callable[..., None]
+UnsubscribeFn = Callable[..., None]
+
+
+class EventEmitter:
+    def __init__(self, bus: 'Bus') -> None:
+        self.bus = bus
+
+    def emit(self, id: str, *args: object):
+        def add_cb(fn: HandlerFn):
+            def cb():
+                r = fn(*args)
+                if r is None:
+                    return
+                # print(r)
+                t = self.bus.loop.create_task(r)
+                self.bus.pending_tasks.append(t)
+                # print(t)
+            self.bus.loop.call_soon(cb)
+
+        self.bus.force_jd_thread()
+        if not hasattr(self, "_listeners"):
+            return
+        idx = 0
+        while idx < len(self._listeners):
+            lid, fn, once = self._listeners[idx]
+            if lid == id:
+                # note that add_cb() can't be inlined here due to lack of block scope in Python
+                add_cb(fn)
+                if once:
+                    del self._listeners[idx]
+                    idx -= 1
+            idx += 1
+
+    def _init_emitter(self):
+        if not hasattr(self, "_listeners"):
+            self._listeners: List[Tuple[str, HandlerFn, bool]] = []
+
+    def on(self, id: str, fn: HandlerFn) -> UnsubscribeFn:
+        """Subscribes an event to a handler. Returns a callback to unsubscribe.
+
+        Args:
+            id (str): event identifier
+            fn (HandlerFn): event callback
+        Returns: callback to unsubscribe
+        """
+        self._init_emitter()
+        self._listeners.append((id, fn, False))
+
+        def unsubscribe():
+            return self.off(id, fn)
+        return unsubscribe
+
+    def once(self, id: str, fn: HandlerFn):
+        """Subscribes an event to run once; then get unsubscribed
+
+        Args:
+            id (str): event identifier
+            fn (HandlerFn): event callback
+        """
+        self._init_emitter()
+        self._listeners.append((id, fn, True))
+
+    def off(self, id: str, fn: HandlerFn):
+        """Unsubscribes a handler from an event
+
+        Args:
+            id (str): event identifier
+            fn (HandlerFn): event callback
+        """
+        self._init_emitter()
+        for i in range(len(self._listeners)):
+            id2, fn2, _ign = self._listeners[i]
+            if id == id2 and fn is fn2:
+                del self._listeners[i]
+                return
+        raise ValueError("no matching on() for off()")
+
+    # usage: await x.event("...")
+    async def event(self, id: str):
+        f = self.bus.loop.create_future()
+        self.once(id, lambda: f.set_result(None))
+        await f
+
+    def wait_for(self, id: str):
+        self.bus.force_non_jd_thread()
+        cv = threading.Condition()
+        happened = False
+
+        def poke(*args: object):
+            nonlocal happened
+            with cv:
+                happened = True
+                cv.notify()
+
+        self.once(id, poke)
+        with cv:
+            while not happened:
+                cv.wait()
+
+    def _log_report_prefix(self) -> str:
+        return ""
+
+    def _add_log_report(self, priority: int, text: str, *args: object):
+        prefix = self._log_report_prefix()
+        msg = prefix + text
+        logger = self.bus.logger
+        if logger:
+            logger.report(priority, msg, *args)
+
+    def log(self, text: str, *args: object):
+        self._add_log_report(JD_LOGGER_PRIORITY_LOG, text, *args)
+
+    def warn(self, text: str, *args: object):
+        self._add_log_report(JD_LOGGER_PRIORITY_WARNING, text, *args)
+
+    def debug(self, text: str, *args: object):
+        self._add_log_report(JD_LOGGER_PRIORITY_DEBUG, text, *args)
+
+    def error(self, text: str, *args: object):
+        self._add_log_report(JD_LOGGER_PRIORITY_ERROR, text, *args)
 
 
 def _service_matches(dev: 'Device', serv: bytearray):
@@ -64,25 +187,53 @@ class Transport:
 # TODO: replace randbytes
 
 
-def _rand_device_id():
-    r: List[int] = []
-    for i in range(8):
-        r.append(random.getrandbits(8))
-    return bytearray(r).hex()
+def rand_u64():
+    return bytearray([getrandbits(8) for _ in range(8)])
 
 
 class Bus(EventEmitter):
     """A Jacdac bus that managed devices, service client, registers."""
 
-    def __init__(self, transport: Transport, *, device_id: str = None) -> None:
+    def __init__(self, transport: Transport, *,
+                 device_id: str = None,
+                 product_identifier: int = None,
+                 firmware_version: str = None,
+                 device_description: str = None,
+                 disable_logger: bool = False,
+                 disable_brain: bool = False,
+                 default_logger_min_priority: int = JD_LOGGER_PRIORITY_SILENT,
+                 settings_file_name: str = None) -> None:
+        """Instantiates a new Jacdac bus
+
+        Args:
+            transport (Transport): packet transport
+            settings_file_name (str): Optional settings file location. Enables settings service.
+            device_id (str, optional): Optional device identifier. Auto-generated if not specified.
+            product_identifier (int, optional): Optional product identifier.
+            device_description (str, optional): Optional device description.
+            disable_logger (bool, optional): Disable the logger service. Defaults to False.
+            disable_brain (bool, optional): Disable unique brain service. Defaults to False.
+            default_logger_min_priority (int, optional): Optional mininimum logger priority. Defaults to JD_LOGGER_PRIORITY_SILENT.
+        """
         super().__init__(self)
         self.devices: List['Device'] = []
         self.unattached_clients: List['Client'] = []
         self.all_clients: List['Client'] = []
         self.servers: List['Server'] = []
+        self.logger: Optional[LoggerServer] = None
+        self.pipes: List['InPipe'] = []
+
+        self.product_identifier = product_identifier
+        self.firmware_version = firmware_version
+        self.device_description = device_description
+        self.disable_brain = disable_brain
+        self.disable_logger = disable_logger
+        self.default_logger_min_priority = default_logger_min_priority
+        self.settings_file_name = settings_file_name
         self._event_counter = 0
+
         if device_id is None:
-            device_id = _rand_device_id()
+            device_id = rand_u64().hex()
         self.self_device = Device(self, device_id, bytearray(4))
         self.process_thread = threading.Thread(target=self._process_task)
         self.transport = transport
@@ -103,7 +254,7 @@ class Bus(EventEmitter):
 
         self.process_thread.start()
 
-        log("starting bus, self={}", self.self_device)
+        self.log("starting bus, self={}", self.self_device)
 
     def run(self, cb: Callable[..., None], *args: Any):
         if self.process_thread is threading.current_thread():
@@ -123,8 +274,15 @@ class Bus(EventEmitter):
         # TODO: what's the best way to import these things
         ctrls = ControlServer(self)  # attach control server
 
-        # TODO: make this optional.
-        brain = UniqueBrainServer(self)
+        if not self.disable_logger:
+            self.logger = LoggerServer(self)
+
+        if self.settings_file_name:
+            from .settings.server import SettingsServer
+            self.settings = SettingsServer(self)
+
+        if not self.disable_brain:
+            UniqueBrainServer(self)
 
         def keep_task(t: 'asyncio.Task[None]'):
             if t.done():
@@ -161,7 +319,7 @@ class Bus(EventEmitter):
             while ptr < 12 + frame[2]:
                 sz = frame[ptr] + 4
                 pktbytes = frame[0:12] + frame[ptr:ptr+sz]
-                #log("PKT: {}-{} / {}", ptr, len(frame), pktbytes.hex())
+                # log("PKT: {}-{} / {}", ptr, len(frame), pktbytes.hex())
                 pkt = JDPacket(frombytes=pktbytes)
                 if ptr > 12:
                     pkt.requires_ack = False
@@ -183,6 +341,12 @@ class Bus(EventEmitter):
             print(info)
         print("END")
 
+    def lookup_server(self, service_class: int) -> Optional['Server']:
+        for s in self.servers:
+            if s.service_class == service_class:
+                return s
+        return None
+
     def _gc_devices(self):
         now_ = now()
         cutoff = now_ - 2000
@@ -226,8 +390,8 @@ class Bus(EventEmitter):
 
     def _reattach(self, dev: 'Device'):
         dev.last_seen = now()
-        log("reattaching services to {}; {}/{} to attach", dev,
-            len(self.unattached_clients), len(self.all_clients))
+        self.debug("reattaching services to {}; {}/{} to attach", dev,
+                   len(self.unattached_clients), len(self.all_clients))
         new_clients: List['Client'] = []
         occupied = bytearray(dev.num_service_classes)
         for c in dev.clients:
@@ -260,27 +424,36 @@ class Bus(EventEmitter):
 
     def process_packet(self, pkt: JDPacket):
         logv("route: {}", pkt)
-        dev_id = pkt.device_identifier
+        dev_id = pkt.device_id
         multi_command_class = pkt.multicommand_class
+        service_index = pkt.service_index
 
         # TODO implement send queue for packet compression
 
-        # if (pkt.requires_ack):
-        #     pkt.requires_ack = False  # make sure we only do it once
-        #     if pkt.device_identifier == self.self_device.device_id:
-        #         ack = JDPacket(cmd=pkt.crc)
-        #         ack.service_index = JD_SERVICE_INDEX_CRC_ACK
-        #         ack._send_report(self.self_device)
+        if pkt.requires_ack and pkt.device_id == self.self_device.device_id:
+            ack = JDPacket(cmd=pkt.crc)
+            ack.service_index = JD_SERVICE_INDEX_CRC_ACK
+            ack.device_id = self.self_device.device_id
+            self._send_core(ack)
 
         self.emit(EV_PACKET_PROCESS, pkt)
 
+        if service_index == JD_SERVICE_INDEX_PIPE and pkt.device_id == self.self_device.device_id:
+            port = pkt.service_command >> PIPE_PORT_SHIFT
+            pipe = next(filter(lambda p: p.port == port, self.pipes), None)
+            if pipe:
+                pipe.handle_packet(pkt)
+            else:
+                self.debug("unknown pipe port {}", port)
+            return
+
         if multi_command_class != None:
             if not pkt.is_command:
                 return  # only commands supported in multi-command
             for h in self.servers:
                 if h.service_class == multi_command_class:
                     # pretend it's directly addressed to us
-                    pkt.device_identifier = self.self_device.device_id
+                    pkt.device_id = self.self_device.device_id
                     pkt.service_index = h.service_index
                     h.handle_packet_outer(pkt)
         elif dev_id == self.self_device.device_id and pkt.is_command:
@@ -303,7 +476,7 @@ class Bus(EventEmitter):
                     if (dev and dev.reset_count > (pkt.data[0] & 0xf)):
                         # if the reset counter went down, it means the device reseted;
                         # treat it as new device
-                        log("device {} resetted", dev)
+                        self.debug("device resetted")
                         self.devices.remove(dev)
                         dev._destroy()
                         dev = None
@@ -311,7 +484,7 @@ class Bus(EventEmitter):
 
                     matches = False
                     if not dev:
-                        dev = Device(self, pkt.device_identifier, pkt.data)
+                        dev = Device(self, pkt.device_id, pkt.data)
                         # ask for uptime
                         # dev.send_ctrl_command(CMD_GET_REG | ControlReg.Uptime)
                         self.emit(EV_DEVICE_CONNECT, dev)
@@ -335,6 +508,128 @@ class Bus(EventEmitter):
 
             dev.process_packet(pkt)
 
+    def add_pipe(self, pipe: 'InPipe'):
+        port = randrange(1, 511)
+        while any(p.port == port for p in self.pipes):
+            port = randrange(1, 511)
+        pipe.port = port
+        self.pipes.append(pipe)
+
+    def remove_pipe(self, pipe: 'InPipe'):
+        if pipe in self.pipes:
+            self.pipes.remove(pipe)
+
+
+class InPipe(EventEmitter):
+    """Incoming pipe"""
+
+    def __init__(self, bus: Bus):
+        super().__init__(bus)
+        self.bus = bus
+        self.port = -1
+        self.next_cnt = 0
+        self.closed = False
+        self.in_q: List[bytearray] = []
+        self.port = -1
+        self.bus.add_pipe(self)
+
+    def open_command(self, cmd: int):
+        return JDPacket.packed(cmd, "b[8] u16 u16", bytearray.fromhex(self.bus.self_device.device_id), self.port, 0)
+
+    def bytes_available(self) -> int:
+        return reduce(lambda x, y: x + len(y), self.in_q, 0)
+
+    def read(self) -> Optional[bytearray]:
+        while True:
+            if len(self.in_q):
+                return self.in_q.pop(0)
+            if self.closed:
+                return None
+            self.wait_for(EV_REPORT_RECEIVE)
+
+    def _close(self):
+        self.closed = True
+        self.bus.remove_pipe(self)
+
+    def close(self):
+        self._close()
+        self.in_q = []
+
+    def meta(self, pkt: JDPacket):
+        pass
+
+    def handle_packet(self, pkt: JDPacket):
+        cmd = pkt.service_command
+        if (cmd & PIPE_COUNTER_MASK) != (self.next_cnt & PIPE_COUNTER_MASK):
+            return
+        self.next_cnt += 1
+        if cmd & PIPE_CLOSE_MASK:
+            self._close()
+        if cmd & PIPE_METADATA_MASK:
+            self.meta(pkt)
+        else:
+            self.in_q.append(pkt.data)
+            self.emit(EV_REPORT_RECEIVE)
+
+    def read_list(self) -> List[bytearray]:
+        r: List[bytearray] = []
+        while True:
+            buf = self.read()
+            if not buf:
+                break
+            if len(buf):
+                r.append(buf)
+        return r
+
+
+class OutPipe(EventEmitter):
+    """Out going packet pipe"""
+
+    def __init__(self, bus: 'Bus', pkt: JDPacket) -> None:
+        super().__init__(bus)
+        [device_id_bytes, port] = pkt.unpack("b[8] u16")
+
+        self.device_id = cast(bytearray, device_id_bytes).hex()
+        self.port = cast(int, port)
+        self.next_cnt = 0
+
+    @property
+    def open(self):
+        return not not self.port
+
+    def write_ex(self, buf: bytearray, flags: int):
+        if not self.port:
+            return
+        pkt = JDPacket(
+            cmd=(self.next_cnt & PIPE_COUNTER_MASK) |
+            (self.port << PIPE_PORT_SHIFT) |
+            flags,
+            data=buf
+        )
+        self.next_cnt += 1
+        if flags & PIPE_CLOSE_MASK:
+            self.port = None
+        pkt.service_index = JD_SERVICE_INDEX_PIPE
+        pkt.requires_ack = True
+        pkt.device_id = self.device_id
+        self.bus._send_core(pkt)
+        # TODO: check acks
+        # if not pkt._send_with_ack(self.device_id):
+        #    self.port = None
+        #    throw "out pipe error: no ACK"
+
+    def write(self, buf: bytearray):
+        self.write_ex(buf, 0)
+
+    def write_and_close(self, buf: bytearray):
+        self.write_ex(buf, PIPE_CLOSE_MASK)
+
+    def close(self):
+        self.write_and_close(bytearray(0))
+
+    def write_meta(self, buf: bytearray):
+        self.write_ex(buf, PIPE_METADATA_MASK)
+
 
 class RawRegisterClient(EventEmitter):
     """A Jacdac register client
@@ -480,10 +775,10 @@ class RawRegisterClient(EventEmitter):
 
 
 class Server(EventEmitter):
-    def __init__(self, bus: Bus, service_class: int) -> None:
+    def __init__(self, bus: Bus, service_class: int, *, instance_name: str = None) -> None:
         super().__init__(bus)
         self.service_class = service_class
-        self.instance_name: Optional[str] = None
+        self.instance_name = instance_name
         self.service_index = None
         self._status_code = 0  # u16, u16
         self.service_index = len(self.bus.servers)
@@ -503,7 +798,7 @@ class Server(EventEmitter):
         if cmd == JD_GET(JD_REG_STATUS_CODE):
             self.handle_status_code(pkt)
         elif cmd == JD_GET(JD_REG_INSTANCE_NAME):
-            self.handle_instance_name(pkt)
+            self._handle_instance_name(pkt)
         else:
             # self.state_updated = False
             self.handle_packet(pkt)
@@ -513,15 +808,19 @@ class Server(EventEmitter):
 
     def send_report(self, pkt: JDPacket):
         pkt.service_index = self.service_index
-        pkt.device_identifier = self.bus.self_device.device_id
+        pkt.device_id = self.bus.self_device.device_id
         self.bus._send_core(pkt)
 
     def send_event(self, event_code: int, data: bytes = None):
         pkt = JDPacket(cmd=self.bus.mk_event_cmd(event_code), data=data)
         def resend(): self.send_report(pkt)
-        resend()
-        self.bus.loop.call_later(0.020, resend)
-        self.bus.loop.call_later(0.100, resend)
+
+        def trisend():
+            resend()
+            self.bus.loop.call_later(0.020, resend)
+            self.bus.loop.call_later(0.100, resend)
+
+        self.bus.run(trisend)
 
     def send_change_event(self):
         self.send_event(JD_EV_CHANGE)
@@ -530,6 +829,9 @@ class Server(EventEmitter):
     def handle_status_code(self, pkt: JDPacket):
         self.handle_reg_u32(pkt, JD_REG_STATUS_CODE, self._status_code)
 
+    def handle_reg_u8(self, pkt: JDPacket, register: int, current: int):
+        return self.handle_reg(pkt, register, "u8", current)
+
     def handle_reg_u32(self, pkt: JDPacket, register: int, current: int):
         return self.handle_reg(pkt, register, "u32", current)
 
@@ -561,14 +863,47 @@ class Server(EventEmitter):
                 current = cast(RegType, v)
         return current
 
-    def handle_instance_name(self, pkt: JDPacket):
+    def _handle_instance_name(self, pkt: JDPacket):
         self.send_report(JDPacket(cmd=pkt.service_command,
                          data=bytearray(self.instance_name or "", "utf-8")))
 
-    def log(self, text: str, *args: object):
-        prefix = "{}.{}>".format(self.bus.self_device,
-                                 self.instance_name or self.service_index)
-        log(prefix + text, *args)
+    def _log_report_prefix(self) -> str:
+        return "{}.{}>".format(self.bus.self_device,
+                               self.instance_name or self.service_index)
+
+
+# TODO: stream sensors
+class SensorServer(Server):
+    def __init__(self, bus: Bus, service_class: int, streaming_interval: int, *, instance_name: str = None, streaming_preferred_interval: int = None) -> None:
+        super().__init__(bus, service_class, instance_name=instance_name)
+        self.streaming_samples = 0
+        self.streaming_preferred_interval: Optional[int] = streaming_preferred_interval
+        self.streaming_interval = streaming_interval
+
+    def handle_packet(self, pkt: JDPacket):
+        cmd = pkt.service_command
+        if cmd == JD_GET(JD_REG_STREAMING_SAMPLES) or cmd == JD_SET(JD_REG_STREAMING_SAMPLES):
+            self._handle_streaming_samples(pkt)
+        elif cmd == JD_GET(JD_REG_STREAMING_INTERVAL) or cmd == JD_SET(JD_REG_STREAMING_SAMPLES):
+            self._handle_streaming_interval(pkt)
+        elif cmd == JD_GET(JD_REG_STREAMING_PREFERRED_INTERVAL):
+            self._handle_streaming_preferred_interval(pkt)
+        super().handle_packet(pkt)
+
+    def _handle_streaming_samples(self, pkt: JDPacket):
+        self.streaming_samples = self.handle_reg_u8(pkt, JD_REG_STREAMING_SAMPLES,
+                                                    self.streaming_samples)
+
+    def _handle_streaming_interval(self, pkt: JDPacket):
+        self.streaming_interval = self.handle_reg_u32(
+            pkt, JD_REG_STREAMING_INTERVAL, self.streaming_interval)
+
+    def _handle_streaming_preferred_interval(self, pkt: JDPacket):
+        if self.streaming_preferred_interval:
+            self.handle_reg_u32(
+                pkt, JD_REG_STREAMING_PREFERRED_INTERVAL, self.streaming_preferred_interval)
+        else:
+            self.send_report(pkt.not_implemented())
 
 
 class ControlServer(Server):
@@ -626,18 +961,84 @@ class ControlServer(Server):
 
     def handle_packet(self, pkt: JDPacket):
         if pkt.is_reg_get:
-            if pkt.reg_code == JD_CONTROL_REG_UPTIME:
+            reg_code = pkt.reg_code
+            if reg_code == JD_CONTROL_REG_UPTIME:
                 self.send_report(JDPacket.packed(
                     JD_GET(JD_CONTROL_REG_UPTIME), "u64",  time.monotonic_ns() // 1000))
+            elif self.bus.product_identifier and reg_code == JD_CONTROL_REG_PRODUCT_IDENTIFIER:
+                self.send_report(JDPacket.packed(
+                    JD_GET(JD_CONTROL_REG_PRODUCT_IDENTIFIER), "u32", self.bus.product_identifier))
+            elif self.bus.firmware_version and reg_code == JD_CONTROL_REG_FIRMWARE_VERSION:
+                self.send_report(JDPacket.packed(
+                    JD_GET(JD_CONTROL_REG_PRODUCT_IDENTIFIER), "s", self.bus.firmware_version))
+            elif reg_code == JD_CONTROL_REG_DEVICE_DESCRIPTION:
+                uname = os.uname()
+                descr = "{}, {}, {}, {}, jacdac {}".format(
+                    self.bus.device_description or "",
+                    uname.nodename, uname.sysname, uname.release, JD_VERSION)
+                self.send_report(JDPacket.packed(
+                    JD_GET(JD_CONTROL_REG_DEVICE_DESCRIPTION), "s", descr))
+            else:
+                self.send_report(pkt.not_implemented())
         else:
             cmd = pkt.service_command
             if cmd == JD_CONTROL_CMD_SERVICES:
                 self.queue_announce()
             elif cmd == JD_CONTROL_CMD_IDENTIFY:
-                self.log("identify")
                 self.bus.emit(EV_IDENTIFY)
             elif cmd == JD_CONTROL_CMD_RESET:
-                sys.exit()  # TODO?
+                # TODO: reset support
+                raise RuntimeError("reset requested")
+            else:
+                self.send_report(pkt.not_implemented())
+
+
+class LoggerServer(Server):
+
+    def __init__(self, bus: Bus) -> None:
+        super().__init__(bus, JD_SERVICE_CLASS_LOGGER)
+        self.min_priority = self.bus.default_logger_min_priority
+        self._last_listener_time = 0
+
+    def handle_packet(self, pkt: JDPacket):
+        self.min_priority = self.handle_reg_u8(
+            pkt, JD_LOGGER_REG_MIN_PRIORITY, self.min_priority)
+        cmd = pkt.service_command
+        if cmd == JD_SET(JD_LOGGER_REG_MIN_PRIORITY):
+            d = cast(int, pkt.unpack("u8")[0])
+            self._last_listener_time = now()
+            if d < self.min_priority:
+                self.min_priority = d
+        return super().handle_packet(pkt)
+
+    def report(self, priority: int, msg: str, *args: object):
+        log(msg, *args)
+        cmd: int = -1
+        if priority == JD_LOGGER_PRIORITY_DEBUG:
+            cmd = JD_LOGGER_CMD_DEBUG
+        elif priority == JD_LOGGER_PRIORITY_LOG:
+            cmd = JD_LOGGER_CMD_LOG
+        elif priority == JD_LOGGER_PRIORITY_WARNING:
+            cmd = JD_LOGGER_CMD_WARN
+        elif priority == JD_LOGGER_CMD_ERROR:
+            cmd = JD_LOGGER_CMD_ERROR
+        else:
+            return
+
+        if now() - self._last_listener_time > JD_LOGGER_LISTENER_TIMEOUT:
+            self._last_listener_time = 0
+            self.min_priority = self.bus.default_logger_min_priority
+
+        if not msg or not self._last_listener_time or priority < self.min_priority:
+            return
+
+        chunks = wrap(msg, JD_SERIAL_MAX_PAYLOAD_SIZE)
+
+        def send_chunks():
+            for chunk in chunks:
+                self.send_report(JDPacket.packed(cmd, "s", chunk))
+
+        self.bus.run(send_chunks)
 
 
 class UniqueBrainServer(Server):
@@ -710,9 +1111,10 @@ class Client(EventEmitter):
         if self.current_device is None:
             return
         pkt.service_index = self.service_index
-        pkt.device_identifier = self.current_device.device_id
+        pkt.device_id = self.current_device.device_id
         pkt._header[3] |= JD_FRAME_FLAG_COMMAND
-        self.bus._send_core(pkt)
+
+        self.bus.run(self.bus._send_core, pkt)
 
     def send_cmd_packed(self, cmd: int, *args: PackType):
         if args is None:
@@ -733,13 +1135,13 @@ class Client(EventEmitter):
             self.device = dev
             self.service_index = service_idx
             self.bus.unattached_clients.remove(self)
-        log("attached {}/{} to client {}", dev, service_idx, self.role)
+        self.debug("attached {}/{} to client {}", dev, service_idx, self.role)
         dev.clients.append(self)
         self.emit(EV_CONNECTED)
         return True
 
     def _detach(self):
-        log("detached {}", self.role)
+        self.debug("detached")
         self.service_index = None
         if not self.broadcast:
             assert self.device
@@ -796,6 +1198,9 @@ class Client(EventEmitter):
                 handler(data)
         return self.on(EV_EVENT, cb)
 
+    def _log_report_prefix(self) -> str:
+        return "{}:{}>".format(self.bus.self_device, self.role)
+
 
 class SensorClient(Client):
     """A client for a sensor service"""
@@ -941,11 +1346,14 @@ class Device(EventEmitter):
         return len(self.services) >> 2
 
     def _destroy(self):
-        log("destroy " + self.short_id)
+        self.debug("destroy")
         for c in self.clients:
             c._detach()
         self.clients = None  # type: ignore
 
+    def _log_report_prefix(self) -> str:
+        return "{}>".format(self.short_id)
+
     def process_packet(self, pkt: JDPacket):
         self.last_seen = now()
         self.emit(EV_PACKET_RECEIVE, pkt)
